;----------------------------------------------------------------------------
;--- This header file is (C)opyright Dennis Bareis 2001. All rights reserved.
;----------------------------------------------------------------------------
;
;     MODULE NAME:   BIN2PPW.H
;
;         $Author:   USER "Dennis"  $
;       $Revision:   1.5  $
;           $Date:   06 Mar 2001 17:42:00  $
;        $Logfile:   C:/DBAREIS/Projects.PVCS/PpwAddOn/bin2ppw/bin2ppw.h.pvcs  $
;
;     DESCRIPTION:   If the "INFILE" macro exists then this is the name of a
;                    file which you want dumped into a text format suitable
;                    for use by PPWIZARD for regeneration of the file with
;                    the "RebuildFile" macro.
;
;                    The "BIN2PPW_REBUILD_FILE" macro is supplied alters the
;                    generated code to recreate the file. The value should
;                    be the name of the file to generate when the generated
;                    code is executed. By default the file is only generated
;                    if it does not exist or has a different size. If unsure
;                    either change setup, pass correct parameters or delete
;                    the files to ensure latest generated.
;
;                    If the "INFILE" macro does not exist then you are
;                    probably simply wishing to make use of the
;                    "RebuildFile" macro to rebuild one or more files.
;
;                    The intent here is to be able to create a single self
;                    contained header file for a product and have it generate
;                    all required files. The "OL_DOC.DH" header file is an
;                    example of this where it creates all GIF and applet
;                    files that are required.
;
;                    If you define "BIN2PPW_C2X" then c2x() storage is used,
;                    this will typically produce larger output but will most
;                    likely be a lot faster (but less readable), you may wish
;                    to try this on very large files.
;
;                    If you define "BIN2PPW_VERIFY" then when the original
;                    data has been rebuilt it is checked against the original
;                    to ensure that no corruption has taken place.
;                    This checks not only for corruption of your data file,
;                    any possible flaws in the encoding/decoding process as
;                    well as any possible incompatibilites in newer PPWIZARD
;                    releases.
;                    As you might expect processing is slowed down by this!
;                    Note that if you find this too slow try the latest
;                    Regina 2.2 beta (it fixes a "slow down" feature).
;
;                    The '@' character is treated as hex in the code below,
;                    this is so that the "#NextId" processing does not
;                    convert any '@@' characters!!!! Be careful. This sort
;                    of problem is detected by the CRC checking.
;----------------------------------------------------------------------------


;--- The version number of this header --------------------------------------
#define VERSION_BIN2PPW    01.065

;--- Default variable name --------------------------------------------------
#define? BIN2PPW_DEFAULT_VAR  D

/*--- Rexx expression for extra characters that need encoding as hex --------*/
#define? BIN2PPW_EXTRA_HEX '"' || "'@"     ;;Both quotes allow use in any string type and '@' to protect from #NextId!! (in case doubled up)


;----------------------------------------------------------------------------
;--- Define macro to rebuild files from binary representation ---------------
;----------------------------------------------------------------------------
#define? BIN2PPW_ALWAYS_MAKE_FILE   N             ;;Y/N
#define  RebuildFile                              \
         #evaluate ^^ ^<$Rexx4RebuildFile {$?}>^
#NextId
#DefineRexx 'Rexx4RebuildFile'
       /*--- Work out name of file (adjust case if required) ----------------*/
       @@File = '{$FILE}'
       if '{$KeepCase="N"}' = 'N' then
           @@File = EnsureFileHasCorrectCase('{$FILE}')

       /*--- Do we wish/need to create the file (or is already existing one OK? ---*/
       <$Rexx4NeedToRebuildFile MakeVar="@@Make" FILE=^{$FILE}^ {$?}>

       /*--- Create the file ------------------------------------------------*/
       if  @@Make = 'Y' then
       do
           /*--- We want/need to remake the file ----------------------------*/
           call Info 'Regenerating: "' || @@File || '"'
           call MustDeleteFile @@File
           @@Rc = charout(@@File, {$DataVar="<$BIN2PPW_DEFAULT_VAR>"})
           call DieIfIoErrorOccurred @@File
           @@Rc = stream(@@File, 'c', 'close');
       end;

       /*--- Make sure we record this output file! --------------------------*/
       call AddOutputFileToDependancyList @@File;
#DefineRexx
#DefineRexx 'Rexx4NeedToRebuildFile'
       /*--- Work out name of file (adjust case if required) ----------------*/
       @@File = '{$FILE}'
       if '{$KeepCase="N"}' = 'N' then
           @@File = EnsureFileHasCorrectCase('{$FILE}')

       /*--- Do we wish/need to create the file (or is already existing one OK? ---*/
       if  stream(@@File, 'c', 'query exists') = '' then
           {$MakeVar} = 'Y';           ;;File does not exist!
       else
       do
           /*--- The file already exists (what does user wish?) -------------*/
           {$MakeVar} = '{$AlwaysMake=^<$BIN2PPW_ALWAYS_MAKE_FILE>^}';

           /*--- If user said don't rebuild do anyway if file size differs ---*/
           if  {$MakeVar} = 'N' then
           do
               /*--- If length not passed then can't check, assume OK -------*/
               @@DataLng = {$DataLngExp=""};
               if  @@DataLng <> '' then
               do
                   /*--- We do know the correct length ----------------------*/
                   if  @@DataLng <> stream(@@File, 'c', 'query size') then
                       {$MakeVar} = 'Y';    ;;File sizes differ!
               end;
           end;
       end;
#DefineRexx


;----------------------------------------------------------------------------
;--- Rexx code to check CRC, dies on error ----------------------------------
;----------------------------------------------------------------------------
#NextId
#DefineRexx 'Rexx4CrcCheck'
       /*--- Calculate CRC now ----------------------------------------------*/
       @@32 = Crc32PrePostConditioning();
       @@32 = UpdateCrc32(@@32, {$DataVar});
       @@32 = Crc32PrePostConditioning(@@32);
       @@32 = Crc32InDisplayableForm(@@32);

       /*--- Die if not correct ---------------------------------------------*/
       if  @@32 \== "{$CRC}" then
           error("CRC32 mismatch on generated data!", "Data Length   = " || AddCommasToDecimalNumber(length({$DataVar})), "Original  CRC = {$CRC}", "Recreated CRC = " || @@32);
#DefineRexx



;--- Some encode/decode definitions -----------------------------------------
#define CODE_SWAPMODE  '~'
#define CODE_REPEAT    '`'
#define CODE_REPEAT3   ''
#define MODE_HEX       'H'
#define MODE_ASCII     'A'
#define ALL_HEX_DIGITS '0123456789ABCDEF'       ;;Lower case 'a'-'f' cause implied ascii mode!


;----------------------------------------------------------------------------
;--- Define Code to decode an encoded string --------------------------------
;----------------------------------------------------------------------------
#NextId
#DefineRexx 'Rexx4Decoding'
   /*--- Get parameter ------------------------------------------------------*/
   @@Coded = {$DataVar="<$BIN2PPW_DEFAULT_VAR>"};

   /*--- We start in ASCII mode ---------------------------------------------*/
   @@Mode = <$MODE_ASCII>

   /*--- Rebuild original string --------------------------------------------*/
   @@Orig = '';
   @@AllCodes = <$CODE_SWAPMODE> || <$CODE_REPEAT> || <$CODE_REPEAT3>;
   do  while @@Coded \== ''
       /*--- More to do -----------------------------------------------------*/
       @@Next1 = left(@@Coded, 1);
       select
           /*++++++++++++++++++++++++++++++++++++++++++++*/
           when pos(@@Next1, @@AllCodes) <> 0 then
           /*++++++++++++++++++++++++++++++++++++++++++++*/
           do
               /*--- Have a code --------------------------------------------*/
               @@Coded = substr(@@Coded, 2);
               select
                   /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
                   when @@Next1 = <$CODE_SWAPMODE> then
                   /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
                   do
                       /*--- Swap mode --------------------------------------*/
                       if  @@Mode = <$MODE_ASCII> then
                           @@Mode = <$MODE_HEX>;
                       else
                           @@Mode = <$MODE_ASCII>;
                   end;
                   /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
                   when @@Next1 = <$CODE_REPEAT3> then
                   /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
                   do
                       /*--- Have a repeat of 3 (repeat last byte) ----------*/
                       @@Orig = @@Orig || copies(right(@@Orig, 1), 3);
                   end;
                   /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
                   otherwise
                   /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
                   do
                       /*--- Have a repeat code (repeat last byte) ----------*/
                       parse var @@Coded @@Run 'x' @@Coded;
                       @@Run  = copies(right(@@Orig, 1), x2d(@@Run));
                       @@Orig = @@Orig || @@Run;
                   end;
               end;
           end;
           /*++++++++++++++++++++++++++++++++++++++++++++*/
           when @@Mode = <$MODE_ASCII> then
           /*++++++++++++++++++++++++++++++++++++++++++++*/
           do
               /*--- In Ascii mode so look for next code --------------------*/
               @@CodePos = verify(@@Coded, @@AllCodes, 'M');
               if  @@CodePos = 0 then
               do
                   /*--- Rest of coded string is actually just ASCII --------*/
                   @@Orig  = @@Orig || @@Coded;
                   @@Coded = '';
               end;
               else
               do
                   /*--- We have a code! Grab all ASCII before it -----------*/
                   @@Orig  = @@Orig || left(@@Coded, @@CodePos-1);
                   @@Coded = substr(@@Coded, @@CodePos);
               end;
           end;
           /*++++++++++++++++++++++++++++++++++++++++++++*/
           otherwise
           /*++++++++++++++++++++++++++++++++++++++++++++*/
           do
               /*--- In HEX mode see if Implied swap to ASCII ---------------*/
               if  pos(@@Next1, <$ALL_HEX_DIGITS>) = 0 then
               do
                   /*--- Found implied HEX mode (have ASCII char) -----------*/
                   @@Mode  = <$MODE_ASCII>;
                   @@Orig  = @@Orig || @@Next1;
                   @@Coded = substr(@@Coded, 2);
               end;
               else
               do
                   /*--- Grab next 2 hex digits to make up 1 byte -----------*/
                   @@Next2 = left(@@Coded, 2);
                   @@Coded = substr(@@Coded, 3);
                   if  datatype(@@Next2, 'X') = 0 then
                       CryAndDie('Found invalid hex characters in byte "' || @@Next2 || '"');
                   else
                   do
                       /*--- Valid hex so add to string ---------------------*/
                       @@Orig  = @@Orig || x2c(@@Next2);
                   end;
               end;
           end;
       end;
   end;

   /*--- If user wanted CRC checking then perform it ------------------------*/
   if  '{$CRC=^^}' <> '' then
   do
       /*--- Want a CRC check -----------------------------------------------*/
       <$Rexx4CrcCheck DataVar=^@@Orig^ Crc=^{$CRC}^>
   end;

   /*--- Return the result --------------------------------------------------*/
   {$RcVar} = @@Orig;
#DefineRexx



;--- User may just want the rebuild functionality ---------------------------
#ifndef  INFILE
         #EOF 1
#endif


;--- Check input file specific and apply defaults ---------------------------
#define? BIN2PPW_BASE_INDENT  0       ;;Indent generated code by this # spaces
#define? BIN2PPW_REXX_INDENT  2       ;;Extra indent for rexx code
#define  BIN2PPW_CHUNK_SIZE   120
#ifdef   BIN2PPW_VERIFY
         #define? BIN2PPW_RECORD_CRC  ;;Generate and record CRC
         #define? BIN2PPW_VERIFY_CRC  ;;Verify regenerated data against original CRC
#endif
#ifdef   BIN2PPW_VERIFY_CRC
         #define? BIN2PPW_RECORD_CRC  ;;Can't verify CRC without recording it!
#endif


;----------------------------------------------------------------------------
;--- Define Code to encode a string into ASCII chars only -------------------
;----------------------------------------------------------------------------
#NextId
#DefineRexx ''
   AsciiHexPrepared = 'N'
#DefineRexx
#option PUSH AllowPack=NO
#DefineRexx 'Rexx4Encoding'
   /*--- Get parameter ------------------------------------------------------*/
   @@Clear = {$DataVar="<$BIN2PPW_DEFAULT_VAR>"};

   /*--- We start in ASCII mode ---------------------------------------------*/
   @@Mode = <$MODE_ASCII>

   /*--- Initialize table which indicates which chars are output as hex etc -*/
   if  AsciiHexPrepared <> 'Y' then
   do
       /*--- Set up characters GENERICALLY ----------------------------------*/
       do  @@I = 0 to 31
           _AH_.@@I = <$MODE_HEX>
       end;
       do  @@I = 32 to 127
           _AH_.@@I = <$MODE_ASCII>
       end;
       do  @@I = 128 to 255
           _AH_.@@I = <$MODE_HEX>
       end;

       /*--- Specify that some specific characters be treated as hex -------*/
       @@Extra = <$CODE_REPEAT> || <$CODE_REPEAT3> || <$CODE_SWAPMODE> || <$BIN2PPW_EXTRA_HEX>;
       do  until @@Extra == ''
           /*--- Get next byte ----------------------------------------------*/
           @@This1 = left(@@Extra, 1);
           @@Extra = substr(@@Extra, 2);

           /*--- Now indicate this should be hex ----------------------------*/
           @AC = c2d(@@This1); _AH_.@AC = <$MODE_HEX>;
       end;

       /*--- Only do once per ppwizard invokation ---------------------------*/
       AsciiHexPrepared = 'Y';
   end;

   /*--- Encode the string --------------------------------------------------*/
   @@Coded    = '';
   @@Last     = '';
   @@ClearLng = length(@@Clear)
   do  @@I = 1 to @@ClearLng
       /*--- Get the next character -----------------------------------------*/
       @@Next1  = substr(@@Clear, @@I, 1);

       /*--- Same as last character? ----------------------------------------*/
       if  @@Next1 \== @@Last then
           @@Last = @@Next1;               ;;Remember so we can detect repeats
       else
       do
           /*--- Have repeated char, quick look for not long enough to bother with ---*/
           #define RUN_MIN   3
           @@J = (@@I + <$RUN_MIN>) - 1;
           if  @@J <= @@ClearLng then
           do
               /*--- OK, enough bytes left to allow a minimum repeat --------*/
               @@This1 = substr(@@Clear, @@J, 1);
               if  @@This1 == @@Last then          ;;Last byte of minimum run also correct byte?
               do
                   /*--- OK looking good, now work out exactly what we have -*/
                   @@Run = 0;
                   do  @@J = @@I to @@ClearLng
                       @@This1 = substr(@@Clear, @@J, 1);
                       if  @@This1 == @@Last then
                           @@Run = @@Run + 1;
                       else
                           leave;              ;;At end of run
                   end;

                   /*--- Not worth it unless at least 'x' extra repeats -------------*/
                   if  @@Run >= <$RUN_MIN> then
                   do
                       /*--- Worth using repeat codes -------------------------------*/
                       if  @@Run == 3 then
                           @@Coded = @@Coded || <$CODE_REPEAT3>;
                       else
                           @@Coded = @@Coded || <$CODE_REPEAT> || d2x(@@Run) || 'x';
                       @@I = @@I + (@@Run - 1);        ;;End of loop will take care of "+1"
                       iterate;
                   end;
               end;
           end;
       end;

       /*--- Is this OK as ASCII? -------------------------------------------*/
       @@Next1C = c2d(@@Next1);
       @@Need   = _AH_.@@Next1C;
       if  @@Mode <> @@Need then
       do
           /*--- Need to change mode, output mode change character? ---------*/
           if  @@Need = <$MODE_HEX> then
               @@Coded = @@Coded || <$CODE_SWAPMODE>;
           else
           do
               /*--- Want to swap to ASCII, can we rely on "implied change"? ---*/
               if  pos(@@Next1, <$ALL_HEX_DIGITS>) <> 0 then
                   @@Coded = @@Coded || <$CODE_SWAPMODE>;    ;;The character could be mistaken for hex digit!
           end;
           @@Mode = @@Need;
       end;

       /*--- Now add the code -----------------------------------------------*/
       if  @@Need = <$MODE_ASCII> then
           @@Coded = @@Coded || @@Next1;         ;;ASCII
       else
           @@Coded = @@Coded || c2x(@@Next1);    ;;HEX
   end;

   /*--- We allow encoded strings to be concatenated ------------------------*/
   if  @@Mode = <$MODE_HEX> then
       @@Coded = @@Coded || <$CODE_SWAPMODE>;    ;;Make sure we end in ASCII mode

   /*--- Return the result --------------------------------------------------*/
   {$RcVar} = @@Coded;
#DefineRexx
#option POP


;----------------------------------------------------------------------------
;--- Generate the output file -----------------------------------------------
;----------------------------------------------------------------------------
#DefineRexx ''
   ;--- Make sure input file exists -----------------------------------------
   InFile   = '<$INFILE>'
   InLength = stream(InFile, 'c', 'query size');
   if  InLength = '' then
       error('The input file "' || InFile || '" does not exist!');

   ;--- By default call output ".H" -----------------------------------------
   #ifdef OUTFILE
          OutFile = '<$OUTFILE>'
   #elseif
          OutFile = _filespec('withoutextn', '<$INFILE>') || '.H';
   #endif
   call Info "Generating: " || OutFile;

   ;--- Make sure we can delete the output file -----------------------------
   call MustDeleteFile OutFile
   call AddOutputFileToDependancyList OutFile;

   ;--- Read the input file -------------------------------------------------
   call AddInputFileToDependancyList InFile;
   CloseRc  = stream(InFile, 'c', 'close');
   Clear   = charin(InFile, 1, InLength);
   call DieIfIoErrorOccurred InFile
   CloseRc = stream(InFile, 'c', 'close');

   /*--- User want data verified? -------------------------------------------*/
   #ifdef  BIN2PPW_RECORD_CRC
           Crc32 = Crc32PrePostConditioning();
           Crc32 = UpdateCrc32(Crc32, Clear);
           Crc32 = Crc32PrePostConditioning(Crc32);
           Crc32 = Crc32InDisplayableForm(Crc32);
   #endif

   ;--- Encode the information ----------------------------------------------
   #ifdef  BIN2PPW_C2X
       Coded     = c2x(Clear);                       ;;Faster but probably bigger output
       ChunkSize = (<$BIN2PPW_CHUNK_SIZE> % 2) * 2;  ;;MUST BE EVEN!
   #elseif
       <$Rexx4Encoding RcVar=^Coded^ DataVar=^Clear^>
       ChunkSize = <$BIN2PPW_CHUNK_SIZE>
   #endif
   OutLength = length(Coded)
   call Info 'Encoded ' || AddCommasToDecimalNumber(InLength) || ' bytes into ' || AddCommasToDecimalNumber(OutLength) || '.'

   ;--- Output header -------------------------------------------------------
   BaseIndent = copies(' ', <$BIN2PPW_BASE_INDENT>);
   RexxIndent = copies(' ', <$BIN2PPW_REXX_INDENT>) || BaseIndent;
   call lineout OutFile, BaseIndent || ";-----------------------------------"
   call lineout OutFile, BaseIndent || ";--- CREATED AT: " || date('N') || ', ' || GetAmPmTime()
   call lineout OutFile, BaseIndent || ";--- GENERATOR : <?InputComponent>"
   call lineout OutFile, BaseIndent || ";--- VERSION   : <$VERSION_BIN2PPW>"
   call lineout OutFile, BaseIndent || ";--- INPUT FILE: " || InFile
   call lineout OutFile, BaseIndent || ";--- FILE TIME : " || stream(InFile, 'c', 'query datetime');
   #ifdef  BIN2PPW_RECORD_CRC
           call lineout OutFile, BaseIndent || ";--- FILE CRC32: " || Crc32;
   #endif
   call lineout OutFile, BaseIndent || ";--- FILE LNG  : " || AddCommasToDecimalNumber(InLength)
   call lineout OutFile, BaseIndent || ";-----------------------------------"
   call lineout OutFile, BaseIndent || "#DefineRexx ''"
   #ifdef  BIN2PPW_REBUILD_FILE
       call lineout OutFile, RexxIndent || "<" || '$Rexx4NeedToRebuildFile MakeVar="RebuildFile" FILE="<$BIN2PPW_REBUILD_FILE>" DataLngExp=^' || InLength || '^>'
       call lineout OutFile, RexxIndent || "if RebuildFile = 'Y' then";
       call lineout OutFile, RexxIndent || "do";
       OldIndent  = RexxIndent;
       RexxIndent = RexxIndent || RexxIndent;
   #endif
   call lineout OutFile, RexxIndent || "<$BIN2PPW_DEFAULT_VAR>=''"

   /*--- Output the data ----------------------------------------------------*/
   Bytes     = OutLength
   do  while   Bytes <> 0
       ;--- How much to grab? -----------------------------------------------
       if  Bytes > ChunkSize then
           Grab = ChunkSize
       else
           Grab = Bytes

       ;--- Process the correct amount --------------------------------------
       #ifdef  BIN2PPW_C2X
               /*--- Just code as HEX literal -------------------------------*/
               call lineout OutFile, RexxIndent || "<$BIN2PPW_DEFAULT_VAR>=<$BIN2PPW_DEFAULT_VAR>||'" || left(Coded, Grab) || "'x"
       #elseif
               /*--- Just save ASCII coded character string -----------------*/
               call lineout OutFile, RexxIndent || "<$BIN2PPW_DEFAULT_VAR>=<$BIN2PPW_DEFAULT_VAR>||'" || left(Coded, Grab) || "'"
       #endif
       Coded = substr(Coded, Grab+1);
       Bytes = Bytes - Grab;
   end;

   /*--- Output the trailer -------------------------------------------------*/
   #ifdef  BIN2PPW_C2X
           #ifdef  BIN2PPW_VERIFY_CRC
                   /*--- User is playing safe and verifying recreated data --*/
                   call lineout OutFile, RexxIndent || "<" || "$Rexx4CrcCheck DataVar=^<$BIN2PPW_DEFAULT_VAR>^ Crc=^" || Crc32 || "^>"
           #endif
   #elseif
           /*--- Decode the ASCII coded character string -------------------*/
           #ifdef BIN2PPW_VERIFY_CRC
                  /*--- Decode AND check CRC --------------------------------*/
                  call lineout OutFile, RexxIndent || "<" || "$Rexx4Decoding RcVar=^<$BIN2PPW_DEFAULT_VAR>^ DataVar=^<$BIN2PPW_DEFAULT_VAR>^ Crc=^" || Crc32 || "^>"
           #elseif
                  /*--- Decode only -----------------------------------------*/
                  call lineout OutFile, RexxIndent || "<" || "$Rexx4Decoding RcVar=^<$BIN2PPW_DEFAULT_VAR>^ DataVar=^<$BIN2PPW_DEFAULT_VAR>^>"
           #endif
   #endif
   #ifdef  BIN2PPW_REBUILD_FILE
           /*--- Generate the file ------------------------------------------*/
           call lineout OutFile, RexxIndent || '<' || '$Rexx4RebuildFile DataVar=^<$BIN2PPW_DEFAULT_VAR>^ FILE="<$BIN2PPW_REBUILD_FILE>" DataLngExp=^' || InLength || '^>'

           /*--- End of conditiona file generation block --------------------*/
           RexxIndent = OldIndent;
           call lineout OutFile, RexxIndent || "end";
   #endif
   call lineout OutFile, BaseIndent || "#DefineRexx"
   call DieIfIoErrorOccurred OutFile
   CloseRc = stream(OutFile, 'c', 'close');
#DefineRexx
